@@ -1,20 +1,22 @@
-if RUBY_VERSION >= '1.9'
-  require 'simplecov'
+require 'simplecov'
 
-  if ENV['TRAVIS']
-    require 'coveralls'
-    Coveralls.wear!
-  end
-
-  SimpleCov.start
+if ENV['TRAVIS']
+  require 'coveralls'
+  Coveralls.wear!
 end
 
+SimpleCov.start
+
 require 'omniauth-saml'
 require 'rack/test'
 require 'rexml/document'
 require 'rexml/xpath'
 require 'base64'
 
+TEST_LOGGER = Logger.new(StringIO.new)
+OneLogin::RubySaml::Logging.logger = TEST_LOGGER
+OmniAuth.config.logger = TEST_LOGGER
+
 RSpec.configure do |config|
   config.include Rack::Test::Methods
 end
